/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UCamControl.h
// Creation   : 07 07 97
// Goal       : declares TCamControl virtual class. A class that defines the
//              method for controlling a mobile (pan/tilt/zoom) camera.
//
//  Nuria Oliver 1997, nuria@media.mit.edu

#ifndef _UCAMCONTROL_H
#define _UCAMCONTROL_H

#include <afxmt.h>
// project includes
#include "UErrors.h"


// Errors
static const short     kErrCameraBusy   = 1960;

class TCamControl
{
public :
	
    virtual ErrorNumber     Init (char *device = "COM1");
    // initialise the control on the camera which is attached to device
    // accessed by file who's name is <device>
	
    virtual void            Dispose ();
    // release camera control ressources
	
    inline
		virtual void            Reset ();
    // reset camera
	
    inline
		virtual void            Home ();
    // move camera to home position
	
    inline
		virtual void            SetZoomSpeed (int speed);
    // set camera zooming speed to <speed>.
    //  2 = fast zoom in
    //  1 = slow zoom in
    //  0 = stop zooming
    // -1 = slow zoom out
    // -2 = fast zoom out
	
    inline
		virtual void            SetPanSpeed (float speed);
    // set camera paning speed to <speed>.
    // Unit is deg/sec, positive is camera left
	
    inline
		virtual void            SetTiltSpeed (float speed);
    // set camera tilting speed to <speed>.
    // Unit is deg/sec, positive is camera up.
	
    inline
		virtual float           GetPanSpeed ();
    // return current camera pan speed
	
    inline
		virtual float           GetTiltSpeed ();
    // return current camera tilt speed
	
    inline
		virtual float           GetZoomPosition ();
    // return current camera absolut zoom position
    // unit is unknown
	
    inline
		virtual float           GetPanPosition ();
    // return current camera absolut pan position
    // unit is degrees, positive is camera left
	
    inline
		virtual float           GetTiltPosition ();
    // return current camera absolut tilt position
    // unit is degrees, positive is up
	
	
	
	
protected :
	
    BOOL            m_bReset;           // reset requested
    BOOL            m_bHome;            // move to home position
    float           m_fPanPos;          // camera pan absolut position
    float           m_fTiltPos;         // camera tilt absolut position
    float           m_fZoomPos;         // camera zoom absolut position
    float           m_fPanSpeed;        // camera desired pan speed
    float           m_fTiltSpeed;       // camera desired tilt speed
    int             m_nZoomSpeed;       // camera desired zoom speed
	
    int             m_nFSMState;        // Child Final State Machine state
    UINT	        m_nParentID;
    volatile 
		BOOL		m_bChildDead;
	CEvent		stopEvent;
    CRITICAL_SECTION     m_pSem;
	
    virtual void    DoReset () = 0;
    virtual void    DoHome () = 0;
    virtual void    DoStop () = 0;
    virtual void    ProcessZoomCommand () = 0;
    virtual void    ProcessSpeedCommand () = 0;
    virtual void    UpdatePanPos () = 0;
    virtual void    UpdateTiltPos () = 0;
    virtual void    UpdateZoomPos () = 0;
	
    void            LowLevelProcess ();
	
    friend UINT     CamControlChildMain (LPVOID userData);
public:
	void GoHome()	{ this->DoHome(); }
	void Stop();
};




//
//    Reset
//

inline
void TCamControl::Reset ()
{
	::EnterCriticalSection( &m_pSem );
	m_bReset = TRUE;
	::LeaveCriticalSection( &m_pSem );
}




//
//    Home
//

inline
void TCamControl::Home ()
{
	::EnterCriticalSection( &m_pSem );
	m_bHome = TRUE;
	::LeaveCriticalSection( &m_pSem );
}




//
//    SetZoomSpeed
//

inline
void TCamControl::SetZoomSpeed (int speed)
{
	::EnterCriticalSection( &m_pSem );
	m_nZoomSpeed = speed;
	::LeaveCriticalSection( &m_pSem );
}




//
//    SetPanSpeed
//

inline
void TCamControl::SetPanSpeed (float speed)
{
	::EnterCriticalSection( &m_pSem );
	m_fPanSpeed = speed;
	::LeaveCriticalSection( &m_pSem );
}



//
//    SetTiltSpeed
//

inline
void TCamControl::SetTiltSpeed (float speed)
{
	::EnterCriticalSection( &m_pSem );
	m_fTiltSpeed = speed;
	::LeaveCriticalSection( &m_pSem );
}




//
//    
//

inline
float TCamControl::GetPanSpeed ()
{
	return m_fPanSpeed;
}




//
//    GetTiltSpeed
//

inline
float TCamControl::GetTiltSpeed ()
{
	return m_fTiltSpeed;
}




//
//    GetZoomPosition
//

inline
float TCamControl::GetZoomPosition ()
{
	return m_fZoomPos;
}




//
//    GetPanPosition
//

inline
float TCamControl::GetPanPosition ()
{
	return m_fPanPos;
}




//
//    GetTiltPosition
//

inline
float TCamControl::GetTiltPosition ()
{
	return m_fTiltPos;
}



#endif