/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UConnex.cpp
// Creation   : 07/1996
// Goal       : defines the TConnex class, see "UConnex.h".
//
// Tab        : no
//  Nuria Oliver 1997, nuria@media.mit.edu


#include "stdafx.h"
// project include
#include "UConnex.h"


////////// static data members initialisation

//
//    Init
//

ErrorNumber TConnex::Init ()
{
	// init data members
	m_pCompoId    = (LPDWORD)NULL;
	m_pDomTab     = (TConnexClass*)NULL;
	m_pBuffer     = (LPVOID)NULL;
	m_nSize       = 0;
	
	return kErrNoErr;
}




//
//    Dispose
//

void TConnex::Dispose ()
{
	if (m_pBuffer != (void*)NULL)
	{
		free (m_pBuffer);
		m_pBuffer = (void*)NULL;
		m_nSize = 0;
	}
	
	m_pCompoId    = (LPDWORD)NULL;
	m_pDomTab     = (TConnexClass*)NULL;
}




//
//    Compute
//

ErrorNumber TConnex::Compute (TBitMap*    iSource,
                              int         iRadius,
                              CRect*      iROI,
                              CRect*      iSupporCRect,
                              LPVOID      iSupportColor,
                              CRect*      oBounding,
                              CRect*      oBoundingPos,
                              TBitMap*    oSupport,
                              TGaussianf* oGaussian,
                              BOOL&       oExists)
{
	ErrorNumber     err = kErrNoErr;
	
	DWORD          width, height, offset;
	
	BYTE           *basePtr, *linePtr, *pixPtr; // bitmap pointers
	DWORD          *slPtr, *spPtr;              // support map pointers
	LPDWORD        neighPtr;
	
	int            i, j, tmp;
	DWORD          currentClass;               // current pixel class number
	DWORD          nbclasses = 0;              // current number of classes
	int            x, y;
	
	err = InitProcess (iSource, iRadius, iROI, width, height, offset, basePtr);
	
	if (err != kErrNoErr)
		return err;
	
	linePtr = basePtr + iRadius * offset + iRadius;
	slPtr   = m_pCompoId + iRadius * width + iRadius;
	
	// process each pixel that has not been initialised to 0
	for (i = height-iRadius; i != 0; i--)
	{
		pixPtr  = linePtr;
		spPtr   = slPtr;
		
		for (j = width-2*iRadius; j != 0; j--)
		{
			if (*pixPtr != 0)
			{
				currentClass = 0;
				
				// look in the upper neighboorhood
				for (y = -iRadius; y < 0; y++)
					for (x = -iRadius; x <= iRadius; x++)
					{
						// compute adress of the neighboor pixel
						neighPtr = spPtr + y*width + x;
						
						ProcessNeighboorPixel  (neighPtr, spPtr, currentClass);
					}
					
					// look in the left neighboorhood
					for (x = -iRadius; x < 0; x++)
					{
						// compute adress of the neighboor pixel
						neighPtr = spPtr + x;
						
						ProcessNeighboorPixel  (neighPtr, spPtr, currentClass);
					}
					
					if (!currentClass)
					{
						// no neighboor pixel, create a new class
						nbclasses++;
						
						m_pDomTab[nbclasses].nIn = 0;
						m_pDomTab[nbclasses].nArea = 1;
						
						// set value in support map for current pixel
						*spPtr = nbclasses;
					}
			}
			// go to next pixel to the right
			pixPtr++;
			spPtr++;
		}
		
		// go to next line
		linePtr += offset;
		slPtr   += width;
	}
	
	if (oExists = (BOOL)(nbclasses != 0))
	{
		// find biggest connex
		tmp = FindBiggest (nbclasses);
		
		err = GenerateOutPut (tmp, width, height, oBounding, oBoundingPos,
			oSupport, iSupporCRect, iSupportColor,
			oGaussian);
	}
	
	return err;
}



////////////////////////////////////////////////////////////////////////////////
							  //                                                                            //
							  //                         protected function members                         //
							  //                                                                            //
							  ////////////////////////////////////////////////////////////////////////////////
							  
							  //
							  //    InitProcess
							  //
							  
inline
ErrorNumber TConnex::InitProcess (TBitMap*    iSource,
                                  int         iRadius,
                                  CRect*      iROI,
                                  DWORD&      oWidth,
                                  DWORD&      oHeight,
                                  DWORD&      oOffset,
                                  LPBYTE      &oBase)
{
	CRect           roi;
	LPBYTE          linePtr;
	LPBYTE          pixPtr;
	unsigned int             tmp1, tmp2, i;
	
	// check source bitmap encoding
	if (((iSource->PixelSize())/8) != 1)
		// can only manage 1 byte per pixel source bitmap
		return kErrCantHandleEncoding;
	
	oOffset   = iSource->LineOffset();
	
	// defines the ROI
	if (iROI == (CRect*)NULL)
		roi = iSource->GetValidRect ();
	else
		roi = *iROI;
	
	// size of the area to process
	oWidth  = roi.right   - roi.left;
	oHeight = roi.bottom  - roi.top;
	
	// compute the max. number of connexes
	tmp1 = iRadius + 1;
	tmp2 = oWidth*oHeight / (tmp1*tmp1) + 
		oWidth / tmp1 +
		oHeight / tmp1 + 2;
	
	tmp1 = tmp2*sizeof(TConnexClass)+oWidth * oHeight * sizeof (DWORD);
	
	if (tmp1 > m_nSize)
	{
		m_nSize = tmp1;
		
		if (m_pBuffer != (void*)NULL)
			free ((void*)m_pBuffer);
		
		m_pBuffer = (void*)malloc (m_nSize);
		
		if (m_pBuffer == (void*)NULL)
		{
			m_nSize = 0;
			return kErrMemFull;
		}
		
		m_pCompoId = (LPDWORD)m_pBuffer;
		m_pDomTab = (TConnexClass*)((LPBYTE)m_pBuffer +
			oWidth * oHeight * sizeof (DWORD));
	}
	
	// initialise the class table
	m_pDomTab[0].nIn = 0;
	
	// clear the support map
	//bzero (m_pCompoId, oWidth * oHeight * sizeof (DWORD));
	memset (m_pCompoId, 0, oWidth * oHeight * sizeof (DWORD));
	
	// Source bitmap initialization:
	oBase = (LPBYTE)(iSource->GetBuffer()) + roi.top * oOffset + roi.left;
	
	// the pixels of the 1st mConnexRadius lines are initialized at 0
	// here is an example on a 7x7 image and with mConnexRadius = 2
	// |0|0|0|0|0|0|0|
	// |0|0|0|0|0|0|0|
	// | | | | | | | |
	// | | | | | | | |
	// | | | | | | | |
	// | | | | | | | |
	// | | | | | | | |
	
	linePtr = oBase;
	for (i = iRadius; i != 0; i--)
	{
		memset (linePtr, 0, oWidth);
		linePtr += oOffset;
	}
	
	// same for the pixels of the 1st and last mConnexRadius columns
	// here is an example on a 7x7 image and with mConnexRadius = 2
	// |0|0| | | |0|0|
	// |0|0| | | |0|0|
	// |0|0| | | |0|0|
	// |0|0| | | |0|0|
	// |0|0| | | |0|0|
	// |0|0| | | |0|0|
	// |0|0| | | |0|0|
	tmp1 = oWidth-iRadius;
	for (i = oHeight-iRadius; i != 0; i--)
	{
		for (pixPtr = linePtr+iRadius-1; pixPtr == linePtr; pixPtr--)
		{
			*pixPtr = 0;
			*(pixPtr + tmp1) = 0;
		}
		linePtr += oOffset;
	}
	
	return kErrNoErr;
								  }

								  
								  
								  
								  
								  //
								  //    Representant
								  //
								  
								  inline
									  void TConnex::Representant (DWORD& ioIndex)
								  {
									  DWORD theIndex;
									  
									  while ((theIndex = m_pDomTab[ioIndex].nIn) != 0)
										  ioIndex = theIndex;
								  }
								  
								  
								  
								  
								  //
								  //    GetRepresentant
								  //
								  
								  inline
									  DWORD TConnex::GetRepresentant (DWORD iIndex)
								  {
									  DWORD theIndex;
									  
									  while ((theIndex = m_pDomTab[iIndex].nIn) != 0)
										  iIndex = theIndex;
									  
									  return iIndex;
								  }
								  
								  
								  
								  
								  //
								  //    ProcessNeighboorPixel
								  //
								  
								  inline
									  void TConnex::ProcessNeighboorPixel  (LPDWORD   neighPtr,
                                      LPDWORD   spPtr,
                                      DWORD&   currentClass)
								  {
									  unsigned int      neighClass;
									  
									  // set neighboor class to its representant
									  Representant (*neighPtr);
									  neighClass = *neighPtr;
									  
									  if (neighClass != 0)
										  // there is a neighboor white pixel
										  
										  if (!currentClass)
										  {
											  // the current pixel has no class allocated (first neighboor pixel)
											  
											  // set the class for the current pixel to be 
											  // the representant of the neighboor class
											  *spPtr = neighClass;
											  currentClass = neighClass;
											  
											  // increment area of representant
											  (m_pDomTab[neighClass].nArea)++;
										  }
										  else
											  if (currentClass != neighClass)
											  {
												  // set the representant of the neighboor class
												  // to be the current pixel class
												  m_pDomTab[neighClass].nIn = (unsigned short) currentClass;
												  
												  // update the area of global representant
												  m_pDomTab[currentClass].nArea += m_pDomTab[neighClass].nArea;
												  
												  // update the class of neighboor pixel to be the
												  // representant
												  *neighPtr = currentClass;
											  }
								  }
								  
								  
								  //
								  //    FindBiggest
								  //
								  
								  inline
									  DWORD TConnex::FindBiggest (DWORD iNbClasses)
								  {
									  int       i;
									  unsigned int       maxArea = m_pDomTab[iNbClasses].nArea;
									  int       maxIndex = iNbClasses;
									  
									  for (i = iNbClasses-1; i != 0; i--)
										  if ((!(m_pDomTab[i].nIn) && (m_pDomTab[i].nArea > maxArea)))
										  {
											  maxArea = m_pDomTab[i].nArea;
											  maxIndex = i;
										  }
										  
										  return maxIndex;
								  }
								  
								  
								  
								  
								  //
								  //    GenerateOutPut
								  //
								  
								  inline
									  ErrorNumber TConnex::GenerateOutPut  (DWORD       classN,
                                      DWORD       iWidth,
                                      DWORD       iHeight,
                                      CRect*      oBounding,
                                      CRect*      oBoundingPos,
                                      TBitMap*    oSupport,
                                      CRect*      iSupporCRect,
                                      LPVOID      iSupportColor,
                                      TGaussianf* oGaussian)
								  {
									  CRect         rect;
									  CRect         posRect;
									  CRect         sRect;
									  DWORD         color = kConnexDefaultColor;
									  BYTE         *linePtr, *pixelPtr;
									  DWORD         offset;
									  DWORD        *supPtr;
									  unsigned int           x, y, yplus1;
									  
									  if (oBounding != (CRect*)NULL)
									  {
										  // Set left,top,right,bottom
										  rect.SetRect( iWidth+2, iHeight+2, 0, 0 );
										  //rect.left     = iWidth+2;
										  //rect.top      = iHeight+2;
										  //rect.right    = 0;
										  //rect.bottom   = 0;
									  }
									  
									  if (oSupport != (TBitMap*)NULL)
									  {
										  // test for pixel encoding
										  
										  if (((oSupport->PixelSize())/8) != 4)
											  return kErrCantHandleEncoding;
										  
										  if (iSupporCRect != (CRect*)NULL)
											  sRect = *iSupporCRect;
										  else
											  sRect = oSupport->GetValidRect ();
										  
										  if (((unsigned)(sRect.right - sRect.left) != iWidth) ||
											  ((unsigned)(sRect.bottom - sRect.top) != iHeight))
											  return kErrWrongSize;
										  
										  if (iSupportColor != (LPVOID)NULL)
											  color = *((LPDWORD)iSupportColor);
										  
										  offset = oSupport->LineOffset();
										  linePtr = (LPBYTE)oSupport->GetBuffer () + sRect.top * offset +
											  sRect.left * 4;
									  }
									  
									  if (oGaussian != (TGaussianf*)NULL)
										  oGaussian->setZero ();
									  
									  supPtr = m_pCompoId;
									  for (y = 0; y < iHeight; y++)
									  {
										  pixelPtr = linePtr;
										  
										  yplus1 = y + 1;
										  
										  for (x = 0; x < iWidth; x++)
										  {
											  if (GetRepresentant(*supPtr) == classN)
											  {
												  if (oBounding != (CRect*)NULL)
												  {
													  // compute bounding box
													  if ((unsigned)rect.top == iHeight + 2)
													  {
														  rect.top      = y;
														  posRect.top   = x;
													  }
													  
													  rect.bottom       = yplus1;
													  posRect.bottom    = x;
													  
													  if ((unsigned) rect.left > x)
													  {
														  rect.left     = x;
														  posRect.left  = y;
													  }
													  
													  if ((unsigned) rect.right < x + 1)
													  {
														  rect.right    = x + 1;
														  posRect.right = y;
													  }
												  }
												  
												  if (oSupport != (TBitMap*)NULL)
													  // create support map
													  *((LPDWORD)pixelPtr) = color;
												  
												  if (oGaussian != (TGaussianf*)NULL)
													  // manage gaussian
													  oGaussian->addSample ((float) x, (float) y);
											  }
											  else
												  if (oSupport != (TBitMap*)NULL)
													  // create support map
													  *((LPDWORD)pixelPtr) = 0;
												  
												  supPtr++;
												  pixelPtr += 4;
										  }
										  
										  linePtr += offset;
									  }
									  
									  if (oGaussian != (TGaussianf*)NULL)
										  oGaussian->getGaussianInv ();
									  
									  if (oBounding != (CRect*)NULL)
										  *oBounding = rect;
									  
									  if (oBoundingPos != (CRect*)NULL)
										  *oBoundingPos = posRect;
									  
									  return kErrNoErr;
}

