/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UFastComp.cpp
//
// Date : 02 07 97
// Goal       : see "UFastComp.h"
//  Nuria Oliver 1997, nuria@media.mit.edu

#include "stdafx.h"
#include <math.h>
#include "UFastComp.h"

//
//    FCCenterAndSize
//
// compute center, width and height from a rect
//
// input :
//    0   left
//    1   top
//    2   right
//    3   bottom
//
//  output : (4 params)
//    4   x
//    5   y
//    6   width
//    7   height

int FCCenterAndSize (CfArray &params)
{
	params[4] = (params[0] + params[2]) / 2;
	params[5] = (params[1] + params[3]) / 2;
	params[6] = params[2]-params[0];
	params[7] = params[3]-params[1];
	
	return 4;
}


//
//    FCBox
//
// compute center, width and height from a rect
//
// input :
//    0   x
//    1   y
//    2   width
//    3   height
//
//  output : (4 params)
//    4   left
//    5   top
//    6   right
//    7   bottom

int FCBox (CfArray &params)
{
	float     midw = params[2] / 2;
	float     midh = params[3] / 2;
	
	params[4] = params[0] - midw;
	params[5] = params[1] - midh;
	params[6] = params[0] + midw;
	params[7] = params[1] + midh;
	
	return 4;
}


//
//    FCTransform
//
// compute coordinates of a point (x, y) in the frame reference,
// given its coordinate in a rotated ans translated reference, defined
// by (eigvect, cx, cy). The angle of the rotated reference is also returned.
//
//  input :
//    0   x center of the rotated ref. in frame ref.
//    1   y center of the rotated ref. in frame ref.
//    2   x coordinate of the eigenvector X in the frame reference
//    3   y coordinate of the eigenvector X in the frame reference
//    4   point x position in rotated reference
//    5   point y position in rotated reference
//
//  output : (3 values)
//    6   point x position in frame reference
//    7   point y position in frame reference
//    8   angle of the rotated ref.

int FCTransform (CfArray &params)
{
    // Compute the orientation of the eigenvector X
    float         angle   = (float) atan2 (params[3], params[2]);
    float         s       = (float) sin( angle );
    float         c       = (float) cos( angle );
	
    // The center with respect to the origin of the image is
    // oCx = x + newCx
    // oCy = y - newCy
    // with newCx     = cx*c-s*cy;    
    //      newCy     = cx*s+c*cy;   
    params[6] = (params[0] + params[4]*c - params[5]*s);
    params[7] = (params[1] + params[4]*s + params[5]*c);
    params[8] = angle;
	
    return 3;
}





//
//    FCTransformMult
//
// same as FCTransform except that it converts a list of points
//
//  input :
//    0   x center of the rotated ref. in frame ref.
//    1   y center of the rotated ref. in frame ref.
//    2   x coordinate of the eigenvector X in the frame reference
//    3   y coordinate of the eigenvector X in the frame reference
//    4   Number of points to transform (n)
//    5   point #0 x position in rotated reference
//    6   point #0 y position in rotated reference
//    ...
//    5+i*2 point #i x position in rotated reference
//    6+i*2 point #i y position in rotated reference
//
//  output : (2*n + 1 values)
//    7+(n-1)*2 angle of the rotated ref.
//    8+(n-1)*2 point #0 x position in frame reference
//    9+(n-1)*2 point #0 y position in frame reference
//    ...
//    8+(n-1)*2+i*2 point #i x position in frame reference
//    9+(n-1)*2+i*2 point #i y position in frame reference


int FCTransformMult (CfArray &params)
{
    // Compute the orientation of the eigenvector X
    float         angle     = (float) atan2 (params[3], params[2]);
    float         s         = (float) sin( angle );
    float         c         = (float) cos( angle );
    float         cx        = params[0];
    float         cy        = params[1];
    int           i, limit  = (int)params[4];
    int           mult, ofs = (limit-1)*2;
    
	
    params[7+ofs] = angle;
    for (i = 0; i < limit; i++)
    {
		mult = i << 1;
		params[8+ofs+mult] = (cx + params[5+mult]*c - params[6+mult]*s);
		params[9+ofs+mult] = (cy + params[5+mult]*s + params[6+mult]*c);
    }
	
    return 2*limit+1;
}





//
//    FCTransformPixel
//
// compute coordinates of a point (x, y) in the frame reference,
// given its coordinate in a rotated ans translated reference, defined
// by (eigvect, cx, cy). The angle of the rotated reference is also returned.
// The distance unit in rotated reference is a function of the angle :
//   d(teta) = 1/MAX(abs (cos (teta)), abs (sin (teta)))
//
//  input :
//    0   x center of the rotated ref. in frame ref.
//    1   y center of the rotated ref. in frame ref.
//    2   x coordinate of the eigenvector X in the frame reference
//    3   y coordinate of the eigenvector X in the frame reference
//    4   point x position in rotated reference
//    5   point y position in rotated reference
//
//  output : (3 values)
//    6   point x position in frame reference
//    7   point y position in frame reference
//    8   angle of the rotated ref.

int FCTransformPixel (CfArray &params)
{
    // Compute the orientation of the X eigenvector
    float         angle   = (float) atan (params[3] / params[2]);
    float         s       = (float) sin( angle );
    float         c       = (float) cos( angle );
    float         f       = (float) (1/max(fabs(s),fabs(c)));
	
    // The center with respect to the origin of the image is
    // oCx = x + newCx
    // oCy = y - newCy
    // with newCx     = cx*c-s*cy;    
    //      newCy     = cx*s+c*cy;   
    params[6] = (params[0] + (params[4]*c - params[5]*s)*f);
    params[7] = (params[1] + (params[4]*s + params[5]*c)*f);
    params[8] = angle;
	
    return 3;
}
