/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UHistogram.h
// Creation   : 02 07 97
// Author     : N.OLIVER
// Goal       : declares the THistogram class, wich stores the histogram of
//              a bitmap. For the moment, only bitmap having encoding
//              Encoding_XBGR_8888 are processed
//
//  Nuria Oliver 1997, nuria@media.mit.edu

#ifndef _UHISTOGRAM_H
#define _UHISTOGRAM_H

#include "stdafx.h"
// project includes
#include "UErrors.h"
#include "UBitMap.h"
#include "UStatistics.h"


const float       kDefMaxDistValue  = (float)5.0;

class THistogram
{
public:
	
    ErrorNumber     Init (DWORD reso, BOOL	prob = FALSE);
    // initialise the Histogram so that it can handles the pixel values
    // coded in <encod> format.
    // The histogram will have <reso> possible values in each (if more than
    // one) dimensions.
    // If <reso>=0 then a resonable default value will be used.
	
    void            Dispose ();
    // disposes the Histogram
	
    DWORD			GetResolution ();
    // returns the resolution of the histogram
    float			GetMaxDist () { return m_fMaxDist; }
    // returns the resolution of the histogram
	
    void			SetResolution (float  &reso) { m_Reso = (DWORD)reso; }
    // set the resolution of the histogram
    void			SetMaxDist (float  &maxDist) { m_fMaxDist = maxDist; }
    // set the resolution of the histogram
	
	
    ErrorNumber     Compute  (	VDib* bitmap, 
								BOOL robust = TRUE,
								BYTE threshold = 0,
								CRect *sourceRect = (CRect *)NULL,
								float maxDist = kDefMaxDistValue);
    // compute the histogram of the Region Of Interest of the <bitmap> TBitMap.
    // if <robust> is TRUE, Compute will uses a 2nd order statistical model,
    // with <maxDist> being the mahalanobis distance from model resulting in a
    // 0 value for the histogram.
    // If <threshold> is greater than 0, the the histogram will be binerised
    // using this treshold
	
    ErrorNumber     Compute  (TGaussianf&		stats,
		BYTE				threshold = 0,
		float				maxDist = kDefMaxDistValue);
    // compute the histogram corresponding to distribution <stats>.
    // <maxDist> is the mahalanobis distance from model resulting in a
    // 0 value for the histogram.
    // If <threshold> is greater than 0, the the histogram will be binerised
    // using this treshold
	
	ErrorNumber		BuildGaussian  (TBitMap*	bitmap, 
		TBitMap*	support,
		LPVOID		supportColor,
		TGaussianf& gaussian, 
		CRect *bitmapRect = (CRect *)NULL,
		CRect *supportRect = (CRect *)NULL);
	
    // compute in <gaussian> the color distribution of the pixels of <bitmap>
    // which position correspond to the pixel having <supportColor> color in
    // the <support> bitmap. If <destRect> != NULL then it specifies the area where
    // to compute the gaussian in the supportMap. If <sourceRect> != NULL  
    // it specifies the area where to compute the gaussian in the original bitmap
	
	void     GetGaussian (TGaussianf& stats);
    // returns the current statistical color model of the histogram
	
    ErrorNumber     Apply  (TBitMap* bitmap, TBitMap* res,
		CRect* bitmapRect = (CRect*)NULL,
		CRect* destRect = (CRect *)NULL);
    // Apply the histogram to <bitmap> TBitMap :
    // <res> is a bitmap which ROI has the same size as <bitmap>'s but is
    // encoded as Encoding_G_8. Each pixel of <res> has the histogram
    // value for the color of the corresponding pixel in <bitmap>. This value
    // is normalised in respect to 255.
    // if <bitmapRect> or <destRect> are not null, then the histogram will only be applied inside
    // these rectangles in the input bitmap and the resulting bitmap respectively
	
	ErrorNumber		ApplyProbMap  (TBitMap* bitmap, 
					   TBitMap* res, 
					   VDib*	resProb,
					   CRect* sourceRect = (CRect *)NULL,
					   CRect* destRect = (CRect *)NULL );
	
	
	ErrorNumber BuildFromStatArray (TGaussianf** array,
		float* weights,
		int nbEntry,
		int nbPositive,
		float maxDist = kDefMaxDistValue );
	
	ErrorNumber BuildFromMixModels (TGaussianf* array,
		float* weights,
		int nbEntry,
		float maxDist = kDefMaxDistValue );
	
	ErrorNumber BuildFrom1MixModel ( TGaussianf* posGauss,
		float* posW,
		int nbPosEntry,
		float maxPosDist,
		BOOL normalize = FALSE );
	
	ErrorNumber BuildFrom2MixModels ( TGaussianf* posGauss,
		float* posW,
		int nbPosEntry,
		float maxPosDist,
		TGaussianf* negGauss,
		float* negW,
		int nbNegEntry );
    
  protected:
	  LPBYTE					m_pValues;			// histogram values
	  LPDWORD					m_pLongValues;		// memory for 32 bit ints
	  float*						m_pProbValues;      // probability values
	  DWORD					m_Reso;				// number of pos. values in each dimensions
	  TGaussianf*				m_pColorModel;
	  // color model for robust mode construction
	  float					m_fMaxDist;		    // distance from model that will be 0 in hist.
	  BOOL					m_bRobust;			// TRUE if the histogram uses robust model
	  BOOL					m_bProb;
	  
	  ErrorNumber			    BuildFromStats (float maxDist, BYTE threshold);
	  // build the histogram from the color model.
};

//
//    GetResolution
//

inline
DWORD THistogram::GetResolution ()
{
	return m_Reso;
}

//
//    Compute
//

inline 
ErrorNumber THistogram::Compute (	TGaussianf& stats,
								 BYTE threshold,
								 float maxDist)
{
	ErrorNumber     err = kErrNoErr;
	
	// copy the statistics into the color model
	m_pColorModel->set (stats);
	
	// compute histogram from model
	err = this->BuildFromStats (maxDist, threshold);
	
	return err;
}

//
//    GetGaussian
//

inline 
void THistogram::GetGaussian (TGaussianf& stats)
{
	stats.set (*m_pColorModel);
}

#endif
