/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UKalman.cpp
// Creation   : 07/1996
// Goal       : declares the TKalman class, see "UKalman.h"
//
// Tab        : no
//  Nuria Oliver 1997, nuria@media.mit.edu

#include "stdafx.h"
// unit include
#include "UKalman.h"

// project includes
#include "UObjAlloc.h"
#include "UTimer.h"

//
//    Init
//

ErrorNumber TKalman::Init ()
{
	ErrorNumber       err = kErrNoErr;
	
	// init data members
	m_bInitialised          = FALSE;
	m_pTimer                = (TTimer*)NULL;
	m_fState[kKalTimInd]    = kKalDefTimeDev[0];
	m_fState[kKalTimInd+1]  = kKalDefTimeDev[1];
	
	m_fState[kKalObsInd]    = kKalDefObsCov[0];
	m_fState[kKalObsInd+1]  = kKalDefObsCov[1];
	m_fState[kKalObsInd+2]  = kKalDefObsCov[2];
	m_fState[kKalObsInd+3]  = kKalDefObsCov[3];
	
	// compute inverse of observation covariance matrix
	this->Invert (m_fState+kKalObsInd, m_fState+kKalObIInd);
	
	err = AllocAndInit (m_pTimer);
	
	// error handling
	if (err != kErrNoErr)
		this->Dispose ();
	
	return err;
}




//
//    Dipose
//

void TKalman::Dispose ()
{
	// dispose the timer
	TestDisposeAndFree (m_pTimer);
}




//
//    Update
//

ErrorNumber TKalman::Update (float* obs)
{
	ErrorNumber     err = kErrNoErr;
	float           y[2];
	
	if (!m_bInitialised)
	{
		m_bInitialised = TRUE;
		
		// copy directly the state
		m_fState[kKalVecInd]    = obs[0];
		m_fState[kKalVecInd+1]  = obs[1];
		
		m_fState[kKalCovInd]    = m_fState[kKalObsInd];
		m_fState[kKalCovInd+1]  = m_fState[kKalObsInd+1];
		m_fState[kKalCovInd+2]  = m_fState[kKalObsInd+2];
		m_fState[kKalCovInd+3]  = m_fState[kKalObsInd+3];
		
		// initialise the timer
		m_pTimer->Gap ();
	}
	else
	{
		// save state in case of singularity
		this->CopyState (m_fState, m_fState+kKalSavInd);
		
		// enlarge covariance based from time
		this->ApplyTimeDeviation ();
		
		////// compute new covariance
		
		// compute inverse of state covariance
		// I <- Inv (C)
		err = this->Invert (m_fState + kKalCovInd,
			m_fState + kKalInvInd);
		
		// add it with inverse observation covariance matrix
		// C <- I + O
		this->AddMat (m_fState + kKalInvInd,
			m_fState + kKalObIInd,
			m_fState + kKalCovInd);
		
		// invert the sum
		// C <- Inv (C)
		err |= this->Invert (m_fState + kKalCovInd,
			m_fState + kKalCovInd);
		
		////// compute the new vector estimation
		
		// compute state contribution
		// X <- I.X
		this->MultVect (m_fState + kKalInvInd,
			m_fState + kKalVecInd,
			m_fState + kKalVecInd);
		
		// compute observation contribution
		// Y <- O.Y
		this->MultVect (m_fState + kKalObIInd, obs, y);
		
		// add the 2
		// X <- X+Y
		this->AddVect  (m_fState + kKalVecInd,
			y,
			m_fState + kKalVecInd);
		
		// compute final estimation
		// X <- C.X
		this->MultVect (m_fState + kKalCovInd,
			m_fState + kKalVecInd,
			m_fState + kKalVecInd);
		
		// error handling
		if (err != kErrNoErr)
			// restore state
			this->CopyState (m_fState+kKalSavInd, m_fState);
	}
	
	return err;  
}




////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                         protected function members                         //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

//
//    ApplyTimeDeviation
//

//inline
void TKalman::ApplyTimeDeviation ()
{
	float   dt      = (float)(m_pTimer->Gap ()) / kTimerNbTiksPerSecond;
	float   dt2     = dt * dt;
	
	m_fState[kKalCovInd]    += dt2 * m_fState[kKalTimInd];
	m_fState[kKalCovInd+3]  += dt2 * m_fState[kKalTimInd+1];
}




//
//    CopyState
//

//inline
void TKalman::CopyState (float* src, float* dst)
{
	*(dst++) = *(src++);
	*(dst++) = *(src++);
	*(dst++) = *(src++);
	*(dst++) = *(src++);
	*(dst++) = *(src++);
	*(dst++) = *(src++);
}



//
//    SetObsCovariance
//

//inline
ErrorNumber TKalman::SetObsCovariance (float* cov)
{
	ErrorNumber   err;
	float*        ptr = m_fState+kKalObsInd;
	
	// compute inverse of observation covariance matrix
	err = this->Invert (cov, m_fState+kKalObIInd);
	
	if (err == kErrNoErr)
	{
		*(ptr++) = *(cov++);
		*(ptr++) = *(cov++);
		*(ptr++) = *(cov++);
		*(ptr++) = *(cov++);
	}
	else
	{
		// restore inverse of old observation covariance matrix
		this->Invert (ptr, m_fState+kKalObIInd);
	}
	
	return err;
}




//
//    GetObsCovariance
//

//inline
void TKalman::GetObsCovariance (float* cov)
{
	float*      ptr = m_fState+kKalObsInd;
	
	*(cov++) = *(ptr++);
	*(cov++) = *(ptr++);
	*(cov++) = *(ptr++);
	*(cov++) = *(ptr++);
}




//
//    SetTimeDeviation
//

//inline
void TKalman::SetTimeDeviation (float* timeDeviation)
{
	m_fState[kKalTimInd]     = timeDeviation[0];
	m_fState[kKalTimInd + 1] = timeDeviation[1];
}




//
//    GetTimeDeviation
//

//inline
void TKalman::GetTimeDeviation (float* timeDeviation)
{
	timeDeviation[0] = m_fState[kKalTimInd];
	timeDeviation[1] = m_fState[kKalTimInd + 1];
}




//
//    Invert
//

//inline
ErrorNumber TKalman::Invert (float* source, float* dest)
{
	float a         = source[0];
	float b         = source[2];
	float c         = source[1];
	float d         = source[3];
	
	float det       = a*d - b*c;
	float invdet;
	
	if (det == 0)
		return kErrSingularity;
	else
	{
		invdet  = 1/det;
		
		dest[0] =  d * invdet;
		dest[1] = -c * invdet;
		dest[2] = -b * invdet;
		dest[3] =  a * invdet;
	}
	
	return kErrNoErr;
}




//
//    AddMat
//

//inline
void TKalman::AddMat (float* a, float* b, float* dest)
{
	dest[0] = a[0]+b[0];
	dest[1] = a[1]+b[1];
	dest[2] = a[2]+b[2];
	dest[3] = a[3]+b[3];
}




//
//    AddVect
//

//inline
void TKalman::AddVect (float* a, float* b, float* dest)
{
	dest[0] = a[0]+b[0];
	dest[1] = a[1]+b[1];
}




//
//    MultVect
//

//inline
void TKalman::MultVect (float* a, float* b, float* dest)
{
	float   tmp = b[0];
	
	dest[0] = a[0]*tmp + a[2]*b[1];
	dest[1] = a[1]*tmp + a[3]*b[1];
}

