/***************************************************************************
 *
 * Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
 * rights reserved.
 *
 * Developed  by the Perceptual Computing Section
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 *
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 **************************************************************************/

///////////////////////////////////////////////////////////////////////////
// Filename	: Vision.cpp
// Created	: Tuesday, June 17th, 1997
// Description: A class to do all the computer vision processing
// 
// Nuria Oliver 1997, nuria@media.mit.edu
///////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Vision.h"
#include "UHistogram.h"
#include "UConnex.h"
#include "UFastComp.h"
#include "ImgWndBB.h"
#include "UKalman.h"
#include "UBoxTracker.h"
#include "HistogramDialog.h"
#include "TrackerDialog.h"
#include "StaticFrameDialog.h"
#include "StaticFrameBBDialog.h"

CVision::CVision():stopEvent(FALSE, FALSE, "Vision Stop Event")
{
	memset((void *)&params,0,sizeof(ThreadParams));
	m_bVisionOn = FALSE;
	m_pGrabber = NULL;
	::InitializeCriticalSection( &m_Sem );
}

CVision::~CVision()
{
	::DeleteCriticalSection( &m_Sem );
}


ErrorNumber CVision::Init( VTciGrabber  *grabber,
											int factor )
{
	ErrorNumber		err = kErrNoErr;
	m_bVisionOn = FALSE;
	m_dFactor = factor;
	m_pGrabber = grabber;

	UINT	dWidth = grabber->Width()/factor;
	UINT	dHeight = grabber->Height()/factor;
	UINT	dPixSize = grabber->PixelSize();

	// Initialize the BMINFO 
	BITMAPINFO_8 reducedBmInfo;
	
	// Initialize standard BITMAPINFO_8 structure
	memset((void *)&reducedBmInfo, 0, sizeof(reducedBmInfo));
	reducedBmInfo.bmiHeader.biSize			= sizeof(BITMAPINFOHEADER);
	reducedBmInfo.bmiHeader.biWidth			= dWidth;
	reducedBmInfo.bmiHeader.biHeight		= dHeight;
	reducedBmInfo.bmiHeader.biPlanes		= 1;
	reducedBmInfo.bmiHeader.biBitCount		= dPixSize;
	reducedBmInfo.bmiHeader.biCompression	= BI_RGB;
    reducedBmInfo.bmiHeader.biSizeImage		= 0;
    reducedBmInfo.bmiHeader.biXPelsPerMeter = 0;
    reducedBmInfo.bmiHeader.biYPelsPerMeter = 0;
    reducedBmInfo.bmiHeader.biClrUsed		= 0;
    reducedBmInfo.bmiHeader.biClrImportant	= 0;

	// Calculate the size of the image
	UINT wWidthBytes = (reducedBmInfo.bmiHeader.biWidth * reducedBmInfo.bmiHeader.biBitCount);
	UINT alignedWidthBytes = ((wWidthBytes+31) & (~31))/ 8;
	
	reducedBmInfo.bmiHeader.biSizeImage = alignedWidthBytes * abs ((int)reducedBmInfo.bmiHeader.biHeight);
	
	// HISTOGRAM
	BITMAPINFO_8 histoBmInfo	= reducedBmInfo;
	histoBmInfo.bmiHeader.biBitCount= 8;
	// Divide by four the size of the histogram pixels because
	// it only has one byte per pixel
	// Be sure that the wWidthBytes is aligned in a word
	UINT  histoWidthBytes = (histoBmInfo.bmiHeader.biWidth * histoBmInfo.bmiHeader.biBitCount);
	UINT  alignedHistoWidthBytes = ((histoWidthBytes+31) & (~31))/8;
	histoBmInfo.bmiHeader.biSizeImage = (alignedHistoWidthBytes) * abs ((int)histoBmInfo.bmiHeader.biHeight);
	//histoBmInfo.bmiHeader.biSizeImage = (histoWidthBytes/8) * abs ((int)reducedBmInfo.bmiHeader.biHeight);
	
	
	// Initialize the bitmaps
	m_bm.Init(reducedBmInfo);
	m_histoBM.Init( histoBmInfo );
	m_connexBM.Init( reducedBmInfo );

	return err;
}

void CVision::StopVision()
{
	if ( m_bVisionOn == FALSE ) {
		return;
	}
	stopEvent.SetEvent();
}

void CVision::StartVision( 	  CStaticFrameBBDialog		*pEstim,
											CHistogramDialog	*pHisto,
											CTrackerDialog		*pTracker,
											CStaticFrameDialog		*pConnex )
{
	ErrorNumber		err = kErrNoErr;
	if ( m_bVisionOn )
		return;
	
	// Initialize the box tracker
    err = pTracker->Init();

	if ( err != kErrNoErr )
		return;
	
	params.pVision = this;
	params.pEstim = pEstim; 
	params.pHisto = pHisto;
	params.pTracker = pTracker;
	params.pConnex = pConnex;

	// Forks the thread that does the computer vision

	pTh = AfxBeginThread(VisionThread, &params );
	m_Thread = pTh->m_hThread;
	m_bVisionOn = TRUE;
}

void CVision::Dispose()
{
	StopVision();
	m_pGrabber = NULL;
}

UINT VisionThread(LPVOID param)
{
	VisionThreadParams *params = (VisionThreadParams *)param;
	CVision	*vision = params->pVision;

	CStaticFrameBBDialog			*pEstimDlg = params->pEstim;
	CStaticFrameDialog			*pConnexDlg = params->pConnex;	
	CHistogramDialog	*pHistoDlg = params->pHisto;

	THistogram *histo		= pHistoDlg->Histo();
	CRITICAL_SECTION	*pSem	= pHistoDlg->Sem();

	CDisplay	*pEstimDisplay = &(pEstimDlg->m_picLiveImg);
	CDisplay	*pConnexDisplay = &(pConnexDlg->m_picLiveImg);
	::EnterCriticalSection( pSem);
	CDisplay	*pHistoDisplay = &(pHistoDlg->m_picProcessedImg);
	::LeaveCriticalSection( pSem );

	// Tracker
	CTrackerDialog			*pTracker = params->pTracker;

	TBoxTracker	*tracker	= pTracker->BoxTracker();
	CRITICAL_SECTION	*pTrackerSem = pTracker->Sem();

	TBitMap	*bm = &(vision->m_bm);
	TBitMap	*histoBM = &(vision->m_histoBM);
	TBitMap	*connexBM = &(vision->m_connexBM);
	VTciGrabber *grabber = vision->m_pGrabber;

	while (!grabber->Grabbing()) {
		if (::WaitForSingleObject(vision->stopEvent.m_hObject, 0) == WAIT_OBJECT_0) {
			//vision->stopEvent.ResetEvent();
			return 0;
		}
	}

	CString					s;
	DWORD                   reso     = 1000;
	BYTE                    threshold = 0;
	float                   maxDist   = 25;
	// nuria osterley's model
	TVector2f               skinMean( 0.490607f, 0.303686f);
	TMatrix3f               skinCov( 0.2477e-03f,-0.0501e-03f,0.1124e-03f);
	// Vasi's model
	//TVector2f               skinMean( 0.470607, 0.303686);
	//TMatrix3f               skinCov( 0.2477e-03,-0.0501e-03,0.1124e-03);
    // This is the inverse of the covariance matrix
	//TMatrix3f               skinInvCov( 4437.56,1978.06,9776.8);
	TGaussianf              skinModel(skinMean,skinCov);

	TConnex                  connex;
	int                      connexRadius = 1;
	// Face variables
	CRect                    bb(0,0,0,0);
	CRect	    			 sbb(0,0,0,0);
	CRect                    tmpBBExtremas(0,0,0,0);
	float				 tmpflWidth, tmpflHeight, tmpflX, tmpflY;
	BOOL                     tmpBoolean, faceFound = TRUE;
	TGaussianf               faceGaussian;
	DWORD                    tmpFaceSupportColor = 0x0000ff00;
	TPointArray			 tmpEigPointsArray;
	float				 tmpFaceEigenValues[2];
	TMatrix4f			 tmpFaceEigenVectorsObject;
	TVector2f			 tmpFaceMeanObject;
	float				 tmpFaceMean[2];
	float				*tmpFaceEigenVectors = tmpFaceEigenVectorsObject.getData();
	CfArray			 tmpFloatArray;
	static int               faceNotFoundCounter = 0;
	
	// Tracker variables
	BOOL                     cameraOn = TRUE;
	BOOL				 zoomOn = TRUE;
	float                   *cameraHome = NULL;
	static UINT          resetStartup;
	const float           homeDelay = 7000;
	const float           resetDelay = 9000;
	static	int				updateCounter = 0;

	// Pointer to save the desired camera home position
	static TTimer            globalTimer;

    	// Kalman filters
    	TKalman                  posKalman, sizKalman, eigValKalman, eigVecKalman;
	TKalman			 eigCenKalman;

	//histo.Init(reso);
	::EnterCriticalSection( pSem );
	histo->Compute( skinModel, pHistoDlg->Threshold(), histo->GetMaxDist() );
	::LeaveCriticalSection( pSem );
	connex.Init();
	
	tmpFloatArray.SetSize(22);
	tmpEigPointsArray.SetSize(4);
	
	// Init the timer
	globalTimer.Init();

	// Set tracker parameters
	tmpBBExtremas.SetRect( 1, 1, bm->Width()-1, bm->Height()-1 ); 
	float		heightRatio		= 0.7f;
	float		p				= 0.9f;
	float		d				= 120.0f;
	float		xThreshold		= 5.0;
	float		yThreshold		= 1.0;
	float		ratioThreshold	= 0.1f;
	float		theoRatio		= 0.7f;
	float		theoRatioThreshold	= 1.0;

//	::EnterCriticalSection( pTrackerSem );
	pTracker->SetParameters(&tmpBBExtremas,
                           heightRatio,
                           p,
                           d,
                           xThreshold,
                           yThreshold,
                           ratioThreshold,
                           theoRatio,
                           theoRatioThreshold); 

	// Set camera and zoom on 
	tracker->SetCameraOnOff(TRUE);
	tracker->SetZoomOnOff(TRUE);
	pTracker->OnUpdateData( 0, 0);
//	::LeaveCriticalSection( pTrackerSem );

	int						 minBBHeight = (int) bm->Height()/4; 
	int						 minBBWidth = (int) (minBBHeight*heightRatio);
	
	// Set resetStartUp
	resetStartup = globalTimer.Time();

	// Init the kalman filters
	posKalman.Init();
	sizKalman.Init();
	eigValKalman.Init();
	eigVecKalman.Init();
	eigCenKalman.Init();
	
	// Set the kalman filters parameters
	//  only the size because the other ones take the default ones
	tmpFloatArray[0] = 20;
	tmpFloatArray[1] = 20;
	posKalman.SetTimeDeviation( tmpFloatArray.GetData() );
	sizKalman.SetTimeDeviation( tmpFloatArray.GetData() );
	eigValKalman.SetTimeDeviation( tmpFloatArray.GetData() );
	eigVecKalman.SetTimeDeviation( tmpFloatArray.GetData() );
	eigCenKalman.SetTimeDeviation( tmpFloatArray.GetData() );
	
    tmpFloatArray[0] = 3;
    tmpFloatArray[1] = 2;
    tmpFloatArray[2] = 2;
    tmpFloatArray[3] = 3;
	
	sizKalman.SetObsCovariance( tmpFloatArray.GetData() );
	eigValKalman.SetObsCovariance( tmpFloatArray.GetData() );
	tmpFloatArray[0] = .05f;
    tmpFloatArray[1] = .01f;
    tmpFloatArray[2] = .01f;
    tmpFloatArray[3] = .05f;
	posKalman.SetObsCovariance( tmpFloatArray.GetData() );
	eigVecKalman.SetObsCovariance( tmpFloatArray.GetData() );
	eigCenKalman.SetObsCovariance( tmpFloatArray.GetData() );

	while (TRUE)	
	{
		if (::WaitForSingleObject(vision->stopEvent.m_hObject, 0) == WAIT_OBJECT_0)
			break;

		if ( !grabber->Grabbing() ) {
			continue;
		}
	
		grabber->DirtyReduce(*bm, vision->Factor());

		::EnterCriticalSection( pSem );
		histo->Apply(bm, histoBM);
		::LeaveCriticalSection( pSem );
		// Compute connectivity on the histogram
		connex.Compute ( histoBM, connexRadius, 
			(CRect*)NULL, (CRect*)NULL,
			(LPVOID)(&tmpFaceSupportColor), 
			&bb, &tmpBBExtremas,
			connexBM, 
			&faceGaussian, 
			tmpBoolean);
		
		// FACE PROCESSING (let's assume for the moment that the face has been found)
		//
		//  Kalman filters for the face
        tmpFloatArray[0] = (float)(bb.left + bb.right) / 2;
        tmpFloatArray[1] = (float)(bb.top + bb.bottom) / 2;
		
		posKalman.Update (tmpFloatArray.GetData(), tmpFloatArray.GetData());
        tmpflX = tmpFloatArray[0];
        tmpflY = tmpFloatArray[1];
		
        tmpFloatArray[0] = (float)(bb.right - bb.left);
        tmpFloatArray[1] = (float)(bb.bottom - bb.top);
		
        sizKalman.Update (tmpFloatArray.GetData(), tmpFloatArray.GetData());
        tmpflWidth  = tmpFloatArray[0];
        tmpflHeight = tmpFloatArray[1];
		
      sbb.left     = (int)(tmpflX - tmpflWidth  / 2 + .5);
      sbb.top      = (int)(tmpflY - tmpflHeight / 2 + .5);
      sbb.right    = (int)(tmpflX + tmpflWidth  / 2 + .5);
      sbb.bottom   = (int)(tmpflY + tmpflHeight / 2 + .5);

		float	tmpRatio = tmpflWidth/tmpflHeight;
		// apply bounding box constraints
		if ((tmpflWidth   < minBBWidth) ||
		    (tmpflHeight  < minBBHeight) ||
		    ( 0.2+theoRatio < tmpRatio) ||
			( theoRatio-.2 > tmpRatio )) {
			faceFound = FALSE;

		} else {
			::EnterCriticalSection( pTrackerSem );
			tracker->SetCameraOnOff (TRUE);
			tracker->SetZoomOnOff (TRUE);
			// Update the display data
			if ( (++updateCounter % 2 ) && (tracker->Update (sbb) == FALSE ) ) {
				faceFound = FALSE;
				updateCounter = 1;
			}
			::LeaveCriticalSection( pTrackerSem );
		}
		
		
		if ( faceFound == TRUE ) 
		{
			//  FACE EIGENVECTORS
			//
#if defined _STATS
			// get the face statistics
			float *p;
			p = faceGaussian.getMeanData();
			s.Format("Mean face: %f %f", p[0], p[1] );
			AfxMessageBox( s );
			p = faceGaussian.getInvCovData(  );
			s.Format("Inv Cov face: %f %f %f",	p[0],p[1],p[2] );
			AfxMessageBox( s );
#endif

			faceGaussian.getEigen (tmpFaceEigenValues, tmpFaceEigenVectors);
			tmpFaceMeanObject = faceGaussian.getMean();
			tmpFaceMean[0]= tmpFaceMeanObject(0);
			tmpFaceMean[1]= tmpFaceMeanObject(1);
			
			eigValKalman.Update (tmpFaceEigenValues, tmpFaceEigenValues);
			eigVecKalman.Update (tmpFaceEigenVectors+2, tmpFaceEigenVectors+2);
			eigCenKalman.Update (tmpFaceMean, tmpFaceMean);
			
			// replace eigenvalues by their square root
			tmpFaceEigenValues[0] = (float) sqrt (fabs (tmpFaceEigenValues[0]));
			tmpFaceEigenValues[1] = (float) sqrt (fabs (tmpFaceEigenValues[1]));
			
			// replace X eigen vector by its smoothed value (computed from Y)
			tmpFaceEigenVectors[0] = tmpFaceEigenVectors[3];
			tmpFaceEigenVectors[1] = -tmpFaceEigenVectors[2];
			
			// prepare a FCTransformMult call
			tmpFloatArray[0]   =  tmpFaceMean[0];
			tmpFloatArray[1]   =  tmpFaceMean[1];
			tmpFloatArray[2]   =  tmpFaceEigenVectors[0];
			tmpFloatArray[3]   =  tmpFaceEigenVectors[1];
			tmpFloatArray[4]   =  4.0;
			tmpFloatArray[5]   =  tmpFaceEigenValues[0];
			tmpFloatArray[6]   =  0.0;
			tmpFloatArray[7]   =  -tmpFaceEigenValues[0];
			tmpFloatArray[8]   =  0.0;
			tmpFloatArray[9]   =  0.0;
			tmpFloatArray[10]  =  tmpFaceEigenValues[1];
			tmpFloatArray[11]  =  0.0;
			tmpFloatArray[12]  =  - tmpFaceEigenValues[1];
			
			FCTransformMult (tmpFloatArray);
			
			tmpEigPointsArray[0] = TPoint((int)tmpFloatArray[14], (int)tmpFloatArray[15]);
			tmpEigPointsArray[1] = TPoint((int)tmpFloatArray[16], (int)tmpFloatArray[17]);
			tmpEigPointsArray[2] = TPoint((int)tmpFloatArray[18], (int)tmpFloatArray[19]);
			tmpEigPointsArray[3] = TPoint((int)tmpFloatArray[20], (int)tmpFloatArray[21]);
			
			// Set the eigenPoints of the window
			pEstimDisplay->SetEigPoints( tmpEigPointsArray );
			
			// Draw the face smoothed bb
			pEstimDisplay->SetRect(sbb);
			
			// check to store camera home position
			if ( cameraHome != NULL ) {
				if ( globalTimer.Time () - resetStartup > homeDelay )
				{
					::EnterCriticalSection( pTrackerSem );
					tracker->GetCameraRadPosition( cameraHome[0], 
						cameraHome[1],	
						cameraHome[2] );
					::LeaveCriticalSection( pTrackerSem );
				}
			}
			
		} // if faceFound == TRUE
		pEstimDisplay->AttachMemToDib(bm->GetBuffer(), FALSE);
		if ( !grabber->Grabbing() ) continue;
		pEstimDisplay->Refresh();
		::EnterCriticalSection( pSem );
		pHistoDisplay->AttachMemToDib(histoBM->GetBuffer(), FALSE);
		::LeaveCriticalSection( pSem );
		if ( !grabber->Grabbing() ) continue;
		pHistoDisplay->Refresh();
		pConnexDisplay->AttachMemToDib(connexBM->GetBuffer(), FALSE);
		if ( !grabber->Grabbing() ) continue;
		pConnexDisplay->Refresh();
		// if not face found, drive camera to canonical position
		if ( faceFound == FALSE ) 
		{
			faceNotFoundCounter++;			
			if ( faceNotFoundCounter == 150) {
				// disconect tracking -> camera moves to home position
				::EnterCriticalSection( pTrackerSem );
				if (cameraOn == TRUE)
					tracker->SetCameraOnOff (FALSE);
				if (zoomOn == TRUE)
					tracker->SetZoomOnOff (FALSE);
				// move camera to home position
				tracker->SetCameraHome();
				tracker->SetZoomWide();
				::LeaveCriticalSection( pTrackerSem );
				faceNotFoundCounter = 0;
			}
			sbb.SetRectEmpty();
			faceFound = TRUE;
		} // 	if ( faceFound == FALSE )
	} // while TRUE

	// dispose the bitmap
	CWinThread	*pMyThread = AfxGetThread();
	SetThreadPriority( pMyThread->m_hThread, THREAD_PRIORITY_TIME_CRITICAL );
	connex.Dispose();
	tmpEigPointsArray.RemoveAll();
	tmpFloatArray.RemoveAll();
	posKalman.Dispose();
	sizKalman.Dispose();
	eigValKalman.Dispose();
	eigVecKalman.Dispose();
	eigCenKalman.Dispose();
	
	globalTimer.Dispose();
	vision->m_bVisionOn = FALSE;
	return 0;
}

